/***************************************************************************
                          imageio.cpp  -  description
                             -------------------
    begin                : Tue Oct 5 1999
    copyright            : (C) 1999 by Lawrence E. Widman
    email                : widman@cardiothink.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include "my_imageio.h"
#include <qimage.h>
#include <qimageio.h>
#include <qpixmap.h>
#include <qpngio.h>

#ifdef X_CLIENT
QPixmap  *pixmap;
#endif /* X_CLIENT */
QImageIO iio;
QImage   image;
int open_file(char *s);
extern int smooth_image( int nored, struct ACTION_LIST *actions, 
			 int actions_num, int shrink );
int  write_file( char *output_filename );
int  insert_label_via_libgd( char *output_filename, char *label_text, 
			     int label_x, int label_y, int size );

int my_imageio::qt_connect (char *input_filename, 
		      int nored,  // nored==TRUE->remove red grid
        	      struct ACTION_LIST *actions, int actions_num,
		      int shrink, int label_num, char **label_text, int *label_x, 
		      int *label_y, int *label_size, char *output_filename){
   // printf("Calling open_file with %s\n", input_filename);
   if( open_file( input_filename ) == EXIT_FAILURE )
     return EXIT_FAILURE;
   if( smooth_image( nored, actions, actions_num, shrink )
       == EXIT_FAILURE )
     return EXIT_FAILURE;
#ifdef X_CLIENT
   pixmap = new QPixmap;
   pixmap = iio.image();   // convert to pixmap
   // insert label here
#endif /* X_CLIENT */
   if( write_file( output_filename ) == EXIT_FAILURE )
     return EXIT_FAILURE;
#ifndef X_CLIENT
   if( label_num > 0 ){
     int  i;

     for( i=0; i<label_num; i++ )
       insert_label_via_libgd( output_filename, label_text[i], 
			       label_x[i], label_y[i], label_size[i] );
   }
#endif /* X_CLIENT */
   return EXIT_SUCCESS;
}

int open_file(char *s){

#ifdef QT_PROFESSIONAL
  qInitImageIO();   // enables JPEG and perhaps others in the future
#else
  qInitPngIO();
#endif /* QT_PROFESSIONAL */

  iio.setFileName( s );
  printf("The input image format is %s\n", iio.imageFormat(s));
  if ( iio.read() == TRUE ){        // ok
     //     printf("Read image OK\n");
     image = iio.image();
     if( image.numColors() == 0 ){ 
        printf("Image is %d by %d pixels and is 32-bit color\n",
			   	image.width(), image.height() );
     }
		 else{ 
        printf("Input image is %d by %d pixels and has %d colors\n",
	       image.width(), image.height(), image.numColors() );
     }
     return EXIT_SUCCESS;
  }                   
  else {
     printf("Didn't read image successfully\n");
     return EXIT_FAILURE;
  }
}

int write_file( char *output_filename ){
  char   format[10];

  strcpy( format, "PNG" );

  /* #ifndef X_CLIENT */
  iio.setFormat( format );
  iio.setFileName( output_filename );
  iio.setImage( image );

  if( iio.write() == TRUE ){
  // #else   /* X_CLIENT */
  //  if( pixmap.save( "foo.png", format ) == TRUE ){
  // #endif  /* X_CLIENT */
    printf("Wrote %s in %s format successfully\n", output_filename, format );
    return EXIT_SUCCESS;
  }
  else {
    printf("Didn't write %s in %s format image\n", output_filename, format );
    return EXIT_FAILURE;
  }
}

// The user has the option of defining NO_LABEL to avoid having to link in libgd.a,
// and of defining X_CLIENT to get the same functionality using QApplication/QPainter
// once someone writes the appropriate code.  This flag could also save time in
// deleting regions, since QPainter provides the bltbit() capability.  So does libgd.a,
// but I'll leave in the slower pixel-by-pixel copy loop above now for simplicity.
#if !defined(X_CLIENT) && !defined(NO_LABEL)
// use libgd.a to insert the desired label.  This approach lets this program
// run in a command-line environment.  If we had used the QApplication-dependent
// approach that comes with QT, this program would run only under a GUI.
// The code here is copied shamelessly from gddemo.c in gd-1.7.3
#include "gd.h"
#include "gdfontg.h"   // giant
#include "gdfontl.h"   // large
#include "gdfontmb.h"  // medium bold
#include "gdfonts.h"   // small font
#include "gdfontt.h"   // tiny font
int  insert_label_via_libgd( char *output_filename, char *label_text,
			      int label_x, int label_y, int size ){

  FILE *fp;                     /* for our input and output files */
  gdImagePtr im;                /* our image */
  int black;                    /* Color indexes */

  /* Try to load the file we just wrote */
  fp = fopen( output_filename, "rb" );
  if (!fp) {
    fprintf(stderr, "Can't load the output file we just wrote, \"%s\",\n"
	    "so can't write the label you requested\n", output_filename);
    return EXIT_FAILURE;
  }

  // load our working image
  im = gdImageCreateFromPng(fp);
  fclose(fp);

  /* First color allocated is background. */
  black = gdImageColorAllocate(im, 0, 0, 0);

  // write text
  switch( size ){
  case 1:
    gdImageString(im, gdFontTiny, label_x, label_y, (unsigned char *)label_text, black);
    break;
  case 2:
    gdImageString(im, gdFontSmall, label_x, label_y, (unsigned char *)label_text, black);
    break;
  case 3:
    gdImageString(im, gdFontMediumBold, label_x, label_y, (unsigned char *)label_text, black);
    break;
  case 4:
    gdImageString(im, gdFontLarge, label_x, label_y, (unsigned char *)label_text, black);
    break;
  case 5:
    gdImageString(im, gdFontGiant, label_x, label_y, (unsigned char *)label_text, black);
    break;
  default:
    printf("Unknown font size %d (valid range 1-5).  Not writing label %s "
	   "at %d, %d.\n", size, label_text, label_x, label_y);
    gdImageDestroy(im);
    return EXIT_FAILURE;
    break;
  }

  /* Make output image interlaced (allows "fade in" in some viewers,
     and in the latest web browsers) */
  gdImageInterlace(im, 1);

  // write out the file
  fp = fopen( output_filename , "wb");
  /* Write PNG */
  gdImagePng(im, fp);
  fclose(fp);
  gdImageDestroy(im);

  printf("Added label \"%s\" to PNG file at %d,%d\n", label_text, label_x, label_y);

  return EXIT_SUCCESS;
}
#endif /* defined(LABEL) && !defined(X_CLIENT) */



