/***************************************************************************
                          main.cpp  -  description
                             -------------------
    begin                : Tue Oct  5 04:41:17 CDT 1999
    copyright            : (C) 1999 by Lawrence E. Widman
    email                : widman@cardiothink.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <qglobal.h>
#include <qapplication.h>
#include "my_imageio.h"

/* defines */
#define SIZEOF_INPUT_FILENAME 40
#define MAX_ACTIONS           20 
#define MAX_LABELS            20 
#define CHECK_CP(AA,SPECIFICATION) \
  if( cp == (char *)NULL ){ \
    printf("You need to specify the region as %s\n", SPECIFICATION); \
    printf("There was an error in the %s specification in %s\n", \
           AA, dup_string); \
    printf("Call this function with no arguments for help on arguments\n"); \
    return EXIT_FAILURE; \
  }
#define DEFAULT_OUTPUT_FILENAME "foo.png"

/* globals */
int process_args( int argc, char** argv );
int add_action_list_item( char action, char *arg_string );
int process_label_argument( char *arg );
void usage( char **argv );
void print_copyright();
struct ACTION_LIST  actions[ MAX_ACTIONS ];  // see header file my_imageio.h
int                 actions_num;             // number of current actions
int                 nored;                   // nored==TRUE->remove red grid
int                 shrink;                  // an integer, defaults to 5
int                 label_num;               // number of labels
char               *label_text[MAX_LABELS];
int                 label_x[MAX_LABELS];
int                 label_y[MAX_LABELS];
int                 label_size[MAX_LABELS];

char input_filename[ SIZEOF_INPUT_FILENAME + 1 ];
char output_filename[ SIZEOF_INPUT_FILENAME + 1 ];

int main(int argc, char** argv){

  my_imageio qt_connect;
#ifdef X_CLIENT
  QApplication a ( argc, argv );   // to enable use of QPixmap
#endif /* X_CLIENT */

  // initializations
  actions_num = 0;
  nored       = FALSE;
  shrink      = 5;        // default
  label_num   = 0;
  strcpy( output_filename, DEFAULT_OUTPUT_FILENAME );

  print_copyright();

  if( process_args( argc, argv ) == EXIT_SUCCESS ) 
    {
      printf("The output file (\"%s\") will be in PNG format.\n",
	     output_filename);
      printf("The red grid will %sbe removed.\n", ( nored ? "" : "not " ));
      if( shrink > 1 )
	printf("The image (after edits) will be shrunken by a factor of %d.\n",
	       shrink);
      if( label_num > 0 ){
	int  i;
	for( i=0; i<label_num; i++ )
	  printf("The image will be labeled with \"%s\" starting at "
		 "x %d, y %d in font size %d\n",
		 label_text[i], label_x[i], label_y[i], label_size[i]);
      }
      qt_connect.qt_connect( &input_filename[0], nored, &actions[0], 
			     actions_num, shrink, label_num, &label_text[0], 
			     &label_x[0], &label_y[0], &label_size[0],
			     &output_filename[0] );
    }

return EXIT_SUCCESS;
}

int process_args( int argc, char** argv ){
  int  i;

  if( argc < 2 ){
    usage( argv );
    return EXIT_FAILURE;
  }
  i = 1;
  while( ( i < argc ) && ( *argv[i] == '-' ) ){
    if( ( strcmp( argv[i], "--nored" ) == 0 ) ||
        ( strcmp( argv[i], "-n" ) == 0 ) ){
      nored = TRUE;
    }
    else
    if( ( strcmp( argv[i], "--shrink" ) == 0 ) ||
        ( strcmp( argv[i], "-s" ) == 0 )     ){
      i++;
      if( i >= argc ){
	printf("Oops, you forgot both the shrink factor and the "
	       "input filename\n");
      }
      else {
	shrink = atoi( argv[i] );
      }
    }
    else
#ifndef NO_LABEL
    if( ( strcmp( argv[i], "--label" ) == 0 ) ||
        ( strcmp( argv[i], "-l" ) == 0 )   ){
      i++;
      if( i >= argc ){
	printf("Oops, you forgot both the label text and the "
	       "input filename\n");
      }
      else {
	process_label_argument( argv[i] );
      }
    }
#endif /* NO_LABEL */
    else
      if( ( strcmp( argv[i], "--delete" ) == 0 ) ||
          ( strcmp( argv[i], "-d" ) == 0 ) ){
      i++;
      if( i >= argc ){
	printf("Oops, you forgot both the arguments for deletion and the "
	       "input filename\n");
      }
      else {
	add_action_list_item( 'D', argv[i] );
      }
      }
    else
      if( ( strcmp( argv[i], "--blank" ) == 0 ) ||
          ( strcmp( argv[i], "-b" ) == 0 ) ){
      i++;
      if( i >= argc ){
	printf("Oops, you forgot both the arguments for blanking and the "
	       "input filename\n");
      }
      else {
	add_action_list_item( 'B', argv[i] );
      }
      }
    else
      if( ( strcmp( argv[i], "--outputfile" ) == 0 ) ||
          ( strcmp( argv[i], "-o" ) == 0 ) ){
      i++;
      if( i >= argc ){
	printf("Oops, you forgot both the output filename and the "
	       "input filename\n");
      }
      else {
	strncpy( output_filename, argv[i], SIZEOF_INPUT_FILENAME );
      }
      }
    else
      if( ( strcmp( argv[i], "-?"     ) == 0 ) ||
          ( strcmp( argv[i], "--help" ) == 0 ) ||
	  ( strcmp( argv[i], "-h"     ) == 0 ) ) {
	usage( argv );
	return EXIT_FAILURE;
      }
    else
      {
	printf("Ignoring unknown option \"%s\"\n", argv[i]);
      }
    i++;
  }

  if( i >= argc ){
    usage( argv );
    printf("Looks like you forgot to specify the name of an input file.\n");
    return EXIT_FAILURE;
  }

  strncpy( input_filename, argv[i], SIZEOF_INPUT_FILENAME );
  printf("fixing to open file \"%s\"\n", input_filename);

  return EXIT_SUCCESS;
}


// we use a fixed array with MAX_ACTIONS slots.  I'm too lazy to set
// up linked-list machinery for the few slots we'll actually use.
int add_action_list_item( char action, char *arg_string ){
  char *cp, *dup_string;

  // in case we need to report an error
  dup_string = strdup( arg_string );

  if( actions_num >= MAX_ACTIONS ){
    printf("Too many actions specified: only are %d currently allowed.\n"
	   "To increase this number, recompile with a higher value for\n"
	   "MAX_ACTIONS.  So, I'm ignoring the command -%c%s\n",
	   MAX_ACTIONS, action, arg_string);
    return EXIT_FAILURE;
  }

  actions[ actions_num ].action = action;
  cp = strtok( arg_string, ":" );
  CHECK_CP("x0","x0:y0:width:height");
  actions[ actions_num ].x0 = atoi( cp );
  cp = strtok( NULL, ":" );
  CHECK_CP("y0","x0:y0:width:height");
  actions[ actions_num ].y0 = atoi( cp );
  cp = strtok( NULL, ":" );
  CHECK_CP("width","x0:y0:width:height");
  actions[ actions_num ].w = atoi( cp );
  cp = strtok( NULL, ":" );
  CHECK_CP("height","x0:y0:width:height");
  actions[ actions_num ].h = atoi( cp );
  
  //  printf("Action %d: %c x0 %d y0 %d w%d h%d\n", 
  //	 actions_num + 1, action, 
  //	 actions[ actions_num ].x0, 
  //	 actions[ actions_num ].y0, 
  //	 actions[ actions_num ].w, 
  //	 actions[ actions_num ].h);

  actions_num++;

  if( dup_string != NULL ) free( dup_string );

  return EXIT_SUCCESS;
}

// fill in the global slots "label_text[i]", "label_x[i]", "label_y[i]",
// "label_size[i]", and then increment "label_num".
int process_label_argument( char *arg_string ){

  char *cp, *dup_string;

  if( label_num >= MAX_LABELS ){
    printf("Too many labels specified: only are %d currently allowed.\n"
	   "To increase this number, recompile with a higher value for\n"
	   "MAX_LABELS.  So, I'm ignoring the command -label %s\n",
	   MAX_LABELS, arg_string);
    return EXIT_FAILURE;
  }

  label_text[ label_num ] = strdup( arg_string );
  dup_string = label_text[ label_num ];         // copy for error messages
  if( strchr( arg_string, ':' ) == (char *)NULL ){
    // we don't have any position or size arguments: it's all text
    label_x[ label_num ] = label_y[ label_num ] = 0;
    label_size[ label_num ] = 4;
    return EXIT_SUCCESS;
    }

  // the contents of label_text are good for error messages
  cp = strtok( arg_string, ":" );
  CHECK_CP("x0","\"x0:y0:size:label text\"");
  label_x[ label_num ] = atoi( cp );
  cp = strtok( NULL, ":" );
  CHECK_CP("y0","\"x0:y0:size:label text\"");
  label_y[ label_num ] = atoi( cp );
  cp = strtok( NULL, ":" );
  CHECK_CP("size","\"x0:y0:size:label text\"");
  label_size[ label_num ] = atoi( cp );
  cp = strtok( NULL, ":" );
  CHECK_CP("label text","\"x0:y0:size:label text\"");
  if( label_text[ label_num ] != (char *)NULL ) free( label_text[ label_num ] );
  label_text[ label_num ] = strdup( cp );

  // increment our counter of labels
  label_num++;

  return EXIT_SUCCESS;
}


void usage( char **argv )
{
     printf("Usage: %s [OPTION]... <inputfilename>\n"
         "  -n, --nored       red grid is removed.\n"
         "  -s, --shrink [N]  output image is shrunken by factor of N in\n"
         "                    each dimension.  Shrinking is done AFTER editing with\n"
	 "                    the -D and -B options.  Default value of N is %d.\n"
         "  -b, --blank x0:y0:w:h   Blank region starting at x0,y0 of size w by h.\n"
         "      The specified region will be set to background color.\n"
         "      This option may be specified several times (max B+D = %d).\n"
         "      This option is applied before deletion.\n"
	 "  -d, --delete x0:y0:w:h   Delete region starting at x0,y0 of size w by h.\n"
         "      Either w or h must equal the image's w or h,\n"
         "      and the image will be shrunken in the other dimension.\n"
         "      This option may be specified several times (max B+D = %d).  The\n"
         "      coordinate system is that of the modified image,\n"
	 "      so be careful with serial deletes!\n"
#ifndef NO_LABEL
	 "  -l, --label [x1:y1:size:]text    Starting at pixel location (x1, y1),\n"
         "      text will be written in specified font size (1, tiny; 2, small;\n"
         "      3, medium bold; 4, large; 5, giant).  Default x1, y1, and size are\n"
         "      0, 0, and 4.  This option may be specified several times\n"
	 "      (current limit: %d). The coordinate system is that of the image\n"
	 "      after all delete-region (D) operations have been carried out.\n"
#endif /* NO_LABEL */
	 "  -o, --outputfile: output filename (maximum length %d).  Default is %s\n"
	 "  -h, --help:       print these instructions\n"
	 "  -?:               print these instructions\n"
#ifndef NO_LABEL
	    , argv[0], shrink, MAX_ACTIONS, MAX_ACTIONS, MAX_LABELS, 
#else
	    , argv[0], shrink, MAX_ACTIONS, MAX_ACTIONS, 
#endif /* NO_LABEL */
	    SIZEOF_INPUT_FILENAME, output_filename);
}

void print_copyright(){
  printf("This is the ecg2png program, version %s.\n"
	 "Copyright (c) CardioThink, Inc. 1999.  All rights reserved.\n"
	 "Contact: http://www.cardiothink.com or bjones@cardiothink.com\n",
	 VERSION);
}
