/***************************************************************************
                          smooth_image.cpp  -  description
                             -------------------
    begin                : Tue Oct 5 1999
    copyright            : (C) 1999 by Lawrence E. Widman
    email                : widman@cardiothink.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <qimage.h>
#include <qimageio.h>
#include "my_imageio.h"
 
extern QImageIO iio;
extern QImage   image;

void TranslateColorMap( int nored );
void ReduceColorInImage();
void ShrinkImage( int factor );
QRgb ModifyColor( QRgb colors_old,
	  int red_op,  int red_old, int green_op, int green_old, 
	  int blue_op, int blue_old, 
          int red_new, int green_new, int blue_new, int *changed_p,
	  int *variable,  char *variable_name, int index, char *call_num );
int ModifyColorOP( int c1, int c2, int op );
int edit_image( struct ACTION_LIST *actions, int actions_num );

// This is the main function that calls all the others.
int smooth_image( int nored, struct ACTION_LIST *actions, 
		  int actions_num, int shrink ){
  //  printf("Translating color map\n");
  TranslateColorMap( nored );
  //  printf("Reducing colors in image...");
  ReduceColorInImage();
  //  printf("done\n");
  //  printf("Editing image (deletes and blanks)...\n");
  if( edit_image( actions, actions_num ) == EXIT_FAILURE )
     return EXIT_FAILURE;
  //  printf("Reducing size of image...");
  ShrinkImage( shrink );
  //  printf("done\n");
  return EXIT_SUCCESS;
}


/*
 *
 *   COLOR MAP MODIFICATION
 *
 */
int        SignalColor, GridColor, BackgroundColor;
int        translate[256];  // one slot for each original color.  The value
                            // in each slot is the translated color (hope-
                            // fully, just one of 3 colors (signal, 
                            // background, and foreground).
QRgb       new_color_map[256];   // the new color map after translation
//int        nored;           // if NULL, we leave red in the image.  
//                            // If TRUE, we remove red 
int        nocolorshrink;   // if TRUE, we don't modify the colormap 

enum {
 NO_OP, EQ_OP, LESS_OP,  LE_OP, GREATER_OP, GE_OP, LAST_OP,
};

/* this macro tests for a change in the color map made by the preceding function call.
   If a change was made, we skip to the end of the loop. */
#define TEST_CONTINUE   { if( changed_p == TRUE )  continue; }

/* collapse color map from up to 256 down to 3 colors: white, red, and black */
/* 10/5/99: Argument "nored": if TRUE, we remove the red grid; if FALSE,
   we keep it. */
void TranslateColorMap( int nored ) {
    int  i, changed_p;

    //    if( nored == TRUE )
    //      printf("Colormap with red grid removed:\n");
    //    else
    //      printf("Colormap with red grid retained:\n");      

    /* these are globals that we use for clearing the background elsewhere */
    SignalColor = GridColor = BackgroundColor = -1;

    /* step through the colors in the colormap.  We apply a series of patterns.
       The first pattern to match dominates the following ones. */
    for (i = 0; i < image.numColors(); i++)
      {
	translate[i] = i;

	/* set up to detect changes with the TEST_CONTINUE macro */
	new_color_map[i] = image.color(i);  // RGB color at index i

        /* set Signal color index (we don't change the colors) */
	new_color_map[i] =
	  ModifyColor( image.color(i),
		       EQ_OP, 255,  EQ_OP, 255,  EQ_OP, 255, 
	                      255,         255,         255,   &changed_p,
		       &BackgroundColor, "ECG background color", i, "0" );
	TEST_CONTINUE;

        /* set Signal color index (we don't change the colors) */
	new_color_map[i] =
	  ModifyColor( image.color(i),
		       EQ_OP,   0,  EQ_OP,   0,  EQ_OP,   0, 
	                        0,           0,           0,   &changed_p,
		       &SignalColor, "ECG signal color", i, "1" );
	TEST_CONTINUE;

#ifdef FOO
        /* set Grid color index (we don't change the colors */
	if( nored == FALSE )
	  {
	    new_color_map[i] =
	      ModifyColor( image.color(i),
			   EQ_OP,  255,  EQ_OP,   0,  EQ_OP,   0, 
			   255,           0,           0,   &changed_p,
			   &GridColor, "ECG grid color", i, "2" );
	    TEST_CONTINUE;
	  }
	else
	  {
	    /* remove red color 5/19/97 */
	    new_color_map[i] =
	      ModifyColor( image.color(i),
			   EQ_OP,  255,  EQ_OP,   0,  EQ_OP,   0, 
			   255,         255,         255,   &changed_p,
			   &GridColor, "ECG grid color", i, "3" );
	    TEST_CONTINUE;
	  }
#endif /* FOO */

	/* enhance partial reds */
	if( nored == FALSE )
	  {
	    new_color_map[i] =
	      ModifyColor( image.color(i),
			   GREATER_OP,  150,  LESS_OP,   150,  LESS_OP,  150, 
			   255,           0,               0,     &changed_p,
			   &GridColor, "ECG grid color", i, "4" );      
	    TEST_CONTINUE;
	  }
	else
	  {
	    /* remove red color 5/19/97 */
	    new_color_map[i] =
	      ModifyColor( image.color(i),
			   GREATER_OP,  150,  EQ_OP,   0,  EQ_OP,   0, 
			   255,         255,         255,   &changed_p,
			   &GridColor, "ECG grid color", i, "5" );
	    TEST_CONTINUE;
	  }

	/* enhance partial white (given that we've picked off red already) */
        new_color_map[i] =
	  ModifyColor( image.color(i),
		       GREATER_OP, 0,  GREATER_OP,   127,  GREATER_OP,   127, 
		              255,     255,                255,   &changed_p,
		       &BackgroundColor, "Background Color", i, "6" );
	TEST_CONTINUE;

	/* enhance partial black */
        new_color_map[i] =
	  ModifyColor( image.color(i),
		       LE_OP,  100,  LE_OP,  100,  LE_OP,  100, 
		                 0,            0,            0,   &changed_p,
		       &SignalColor, "ECG signal Color", i, "7" );
	TEST_CONTINUE;

	/* enhance partial white (huh?) */
        new_color_map[i] =
	  ModifyColor( image.color(i),
		       LE_OP, 255,  GE_OP,   0,  GE_OP,   0, 
		              255,         255,         255,   &changed_p,
		       &BackgroundColor, "Background Color", i, "8" );
	TEST_CONTINUE;
      }

    // now finalize any changes in this color
    for (i = 0; i < image.numColors(); i++)
	image.setColor( i, new_color_map[i] );
  }

/* Go through each pixel in the image, and change its index to 0, 1, or 2
   depending on whether its new color is SignalColor, BackgroundColor, or
   GridColor. */
void ReduceColorInImage(){
  
  int  x, y;
  unsigned int  black, red, white;

  black = qRgb(   0,   0,   0 );
  white = qRgb( 255, 255, 255 );
  red   = qRgb( 255,   0,   0 );

  // go through and change all the indices
  for( x=0; x < image.width(); x++ )
    for( y=0; y < image.height(); y++ ){
      if( *(image.scanLine(y) + x) >= image.numColors() ){
	printf("Illegal color index %d at x %d, y %d\n",
	       *(image.scanLine(y) + x), x, y);
      }
      else {
	// signal color
	if( new_color_map[ *(image.scanLine(y) + x) ] == black ){
	  *(image.scanLine(y) + x) = 0;
	}
	else{
	  // background color
	  if( new_color_map[ *(image.scanLine(y) + x) ] == white ){
	    *(image.scanLine(y) + x) = 1;
	  }
	  else{
	    // grid color with qt 1.x
	    if( new_color_map[ *(image.scanLine(y) + x) ] == red ){
	      *(image.scanLine(y) + x) = 2;
	    }
	    else{
	      printf("Unknown translated color %X (index %d) at x %d, y %d\n",
		     new_color_map[*(image.scanLine(y) + x)],
		     *(image.scanLine(y) + x), x, y);
	      *(image.scanLine(y) + x) = 1;
	    }
	  }
	}
      }
    }
  for (x = 0; x < image.numColors(); x++)
    image.setColor( x, black );
  image.setColor( 0, black );     // signal
  image.setColor( 1, white );     // background
  image.setColor( 2, red   );     // grid color
}

/* reduce the size of the image by, say, 4-fold in each dimension */
void ShrinkImage( int factor ){

  int  x, y, i, j, x_new, y_new, w_new, h_new, depth, num_colors;
  int  num_red_cells;
  static QImage smaller_image;
  unsigned int Result;

  if( factor == 1 ) return;  // nothing to do if factor is 1

  w_new = 1 + image.width() / factor;
  h_new = 1 + image.height() / factor;
  depth = 8;    num_colors = 3;
  smaller_image.create( w_new, h_new, depth, num_colors );

  if( smaller_image.isNull() == TRUE ){
    printf("Can't allocate memory for smaller image\n");
    return;
  }

  //  printf("from %d by %d to %d by %d pixels...", image.width(),
  //	 image.height(), w_new, h_new);

  // go through and set new pixel based on grid of factorxfactor old pixels
  // highest priority is the signal, then the grid, then background
  for( x=0, x_new=0; x < image.width()-factor; x=x+factor, x_new++ )
    for( y=0, y_new=0; y < image.height()-factor; y=y+factor, y_new++ ){
      Result = 1;                  // background
      num_red_cells = 0;           // initialize
      for( i=0; i<factor; i++ )
	for( j=0; j<factor; j++){
	  if( *(image.scanLine(y + j) + x + i) > 2 ){
	    printf("Illegal color index %d at x %d, y %d\n",
		   *(image.scanLine(y + j) + x + i), x, y);
	  }
	  // if this is a signal color, we're done scanning this 
	  // factorxfactor cell
	  if( *(image.scanLine(y + j) + x + i) == 0 ){	  
	    Result = 0;
	    break;
	  }
	  // if this is a grid color, we need to keep scanning
	  if( ( *(image.scanLine(y + j) + x + i) == 2 ) && ( Result > 0 ) ){
	    Result = 2;
	    num_red_cells++;
	  }
      }

      // We don't consider this a grid line unless we have a column of
      // red that is 2 pixels wide.
      if( ( Result == 2 ) && ( num_red_cells < (int)(1.5 * factor) ) )
	Result = 1;

      *(smaller_image.scanLine(y_new) + x_new) = Result;
    }

  for( i=0; i<3; i++ ){
    smaller_image.setColor( i, image.color( i ) );
    // printf("setting smaller_image color %d to %X\n", i, image.color( i ));
  }

  image = smaller_image;
}


/* look for old RGB pattern specified by *_old and *_op (operators).
   Thus, if red_old is 0 and red_op is EQ_OP, the matching pattern
   is (current red value) == 0.
     This function substitutes one color for another, and is useful
   for enhancing dithered colors and for removing selected colors.
     Arguments:  pattern for RBG colors, 
                 old RGB colors, 
                 operators to be used for pattern matching,
                 new RGB colors,
		 pointer to flag, set to FALSE if we didn't match the pattern
		   and TRUE if we did,
                 optional variable (SignalColor, GridColor, BackgroundColor) to
                   be updated if the pattern matches,
                 name of variable, used in printing message to user,
                 index in translate[] of the color we're working on here.
 */
QRgb ModifyColor( QRgb colors_old,
	  int red_op,  int red_old, int green_op, int green_old, 
	  int blue_op, int blue_old, 
          int red_new, int green_new, int blue_new, int *changed_p,
	  int *variable,  char *variable_name, int index, char *call_num )
  {

    QRgb colors_new;

    colors_new = colors_old;
    *changed_p = FALSE;

    if ( ModifyColorOP( qRed( colors_old ),   red_old,   red_op )   &&
	 ModifyColorOP( qGreen( colors_old ), green_old, green_op ) &&
	 ModifyColorOP( qBlue( colors_old ),  blue_old,  blue_op ) )
      {

	if( nocolorshrink == FALSE )
	  {

	    //  printf("   Changing %3d %3d %3d to %3d %3d %3d (pattern %s)\n",
	    //	    qRed(colors_old), qGreen(colors_old), qBlue(colors_old),
	    //	    red_new, green_new, blue_new, call_num);
	    colors_new = qRgb( red_new, green_new, blue_new );
	  }
	
	if( variable != (int *)NULL )
	  {
	    if ( *variable < 0 )
	      {
		*variable = index;
		//	printf("The %s index is %d (R%d G%d B%d)\n", 
		//		    variable_name, index, 
		//		    red_new, green_new, blue_new );
	      }
	    else
	      translate[ index ] = *variable;
	  }

	/* tell 'em we matched this pattern */
	*changed_p = TRUE;

      }

    return colors_new;

  }


/* return 1 if c1 has relation "op" to c2, 0 otherwise */
int ModifyColorOP( int c1, int c2, int op )
  {
    int result;

    switch( op )
      {
      case EQ_OP:      result = ( c1 == c2 ); break;
      case LESS_OP:    result = ( c1 <  c2 ); break;
      case LE_OP:      result = ( c1 <= c2 ); break;
      case GREATER_OP: result = ( c1 >  c2 ); break;
      case GE_OP:      result = ( c1 >= c2 ); break; 
      default: 
         printf("Unknown operation %d\n", op ); exit(1); break;
      }

    return( result );
  }

// Perform the region blanking ('B') and deleting ('D') operations.
// If someone wants to do the work of using bltbit() under QT, the
// code to change is in case 'D' below.  This change would limit the
// program to running under windowing systems only (i.e., not from
// the command line when a window system is not running).  The changed
// code should therefore be enclosed within #ifdef X_CLIENT / #endif
// preprocessor statements.
int edit_image( struct ACTION_LIST *actions, int actions_num ){
   int   i;
   int   x, y;
   int   x_start, x_end, y_start, y_end, width, height;
   int   depth, num_colors;
   char  action;
   QImage new_image;

   for( i=0; i < actions_num; i++ ){
     action  = actions[i].action;
     x_start = actions[i].x0;  y_start = actions[i].y0;
     width   = actions[i].w;   height  = actions[i].h;
     x_end   = x_start + width - 1;
     y_end   = y_start + height - 1;

     printf("Action %s at x0 %d y0 %d w %d h %d\n",
  	    ( action == 'D' ? "Delete block" : "Blank(erase) block" ), 
	    x_start, y_start, width, height);

     switch( action ){
     case 'B':
       for (x=x_start; x<=x_end; x++) {
	 for (y=y_start; y<=y_end; y++) {
	   // 1 is the index of the background color
	   *(image.scanLine(y) + x) = 1;     
	 }
       }
       break;
     case 'D':
       depth = 8;   num_colors = 3;
       if( width == image.width() ){
	 new_image.create( width, image.height() - height, depth, num_colors );
	 if( new_image.isNull() == TRUE ){
	   printf("Can't allocate memory for smaller image\n");
	   return EXIT_FAILURE;
	 }
	 // we can't specify subregions in the target, "new_image".  So,
	 // we copy the desired number of pixels from the start of the image
	 // and overwrite by hand with the pixels on the other side of the
	 // deleted region.
	 new_image = image.copy( 0, 0, width, image.height() - height );
	 // new_image = image.copy( 0, y_start+height, width, image.height() );
	 for (x=0; x<width; x++) {
	   for (y=y_start+height; y < image.height(); y++) {
	     // 1 is the index of the background color
	     *(new_image.scanLine(y - height) + x) = 
	       *(image.scanLine(y) + x);
	   }
	 }
	 //	 printf("Image after deletion is %d by %d pixels, "
	 //		"should be %d by %d\n",
	 //		new_image.width(), new_image.height(), 
	 //		image.width(), image.height() - height);
	 iio.setImage( new_image );
	 image = iio.image();
       }
       else
	 if( height == image.height() ){
	   new_image.create( image.width() - width, height, 
			     depth, num_colors );
	   if( new_image.isNull() == TRUE ){
	     printf("Can't allocate memory for smaller image\n");
	     return EXIT_FAILURE;
	   }
	   // we can't specify subregions in the target, "new_image".  So,
	   // we copy the desired number of pixels from the start of the image
	   // and overwrite by hand with the pixels on the other side of the
	   // deleted region.
	   new_image = image.copy( 0, 0, image.width() - width, height );
	   for (x=x_start+width; x < image.width(); x++) {
	     for (y=0; y < height; y++) {
	       // 1 is the index of the background color
	       *(new_image.scanLine(y) + x - width) = 
		 *(image.scanLine(y) + x);
		 }
	   }
	   iio.setImage( new_image );
	   image = iio.image();
	 }
	 else{
	   printf( "Error.  The image dimensions are w %d by h %d.  The "
		   "deleted\nregion must share one of these values\n.", 
		   image.width(), image.height() );
	 }
       break;
     }
   }

   return EXIT_SUCCESS;
}
